@@ -33,6 +33,7 @@
 #include <sys/types.h>
 #include <sys/stat.h>
 #include <ctype.h>
+#include <math.h>
 /* }}} */
 /* {{{ [fold] Structures and types */
 #include "videodev2.h"
@@ -45,6 +46,7 @@ typedef unsigned char Bool;
 #define MAX(a,b)	((a)>(b)?(a):(b))
 #define MIN(a,b)	((a)<(b)?(a):(b))
 #define SIZE(x)		(sizeof(x)/sizeof((x)[0]))
+#define CLIP(a,low,high) MAX((low),MIN((high),(a)))
 
 char *progname = "qcset";
 
@@ -64,6 +66,7 @@ const struct bitflags flags_debug[] = {
 	{ "interrupts",		QC_DEBUGINTERRUPTS },
 	{ "mutex",		QC_DEBUGMUTEX },
 	{ "common",		QC_DEBUGCOMMON },
+	{ "frame",		QC_DEBUGFRAME },
 	{ "all",		QC_DEBUGALL },
 	{ NULL,			0 }
 };
@@ -77,7 +80,9 @@ const struct bitflags flags_quality[] = {
 	{ "fastest",		0 },
 	{ "good",		1 },
 	{ "horizontal",		2 },
-	{ "best",		3 },
+	{ "bilinear",		3 },
+	{ "better",		4 },
+	{ "best",		5 },
 	{ NULL,			0 }
 };
 const struct bitflags flags_bool[] = {
@@ -93,15 +98,17 @@ const struct ioctlstruct {
 	Bool bitfield;
 	const struct bitflags *flags;
 } ioctlinfo[] = {
-	{ "debug",		VIDIOCQCGDEBUG,		VIDIOCQCSDEBUG,		TRUE,	flags_debug },
+	{ "debug",		VIDIOCQCGDEBUG,		VIDIOCQCSDEBUG,		TRUE,	flags_debug },	/* For backwards compatibility */
+	{ "qcdebug",		VIDIOCQCGDEBUG,		VIDIOCQCSDEBUG,		TRUE,	flags_debug },
 	{ "keepsettings",	VIDIOCQCGKEEPSETTINGS,	VIDIOCQCSKEEPSETTINGS,	FALSE,	flags_bool },
-	{ "settle", 		VIDIOCQCGSETTLE,	VIDIOCQCSSETTLE,	FALSE,	flags_bool },
+	{ "settle", 		VIDIOCQCGSETTLE,	VIDIOCQCSSETTLE,	FALSE,	NULL },
 	{ "subsample",		VIDIOCQCGSUBSAMPLE,	VIDIOCQCSSUBSAMPLE,	FALSE,	flags_bool },
 	{ "compress",		VIDIOCQCGCOMPRESS,	VIDIOCQCSCOMPRESS,	FALSE,	flags_bool },
 	{ "frameskip",		VIDIOCQCGFRAMESKIP,	VIDIOCQCSFRAMESKIP,	FALSE,	NULL },
 	{ "quality",		VIDIOCQCGQUALITY,	VIDIOCQCSQUALITY,	FALSE,	flags_quality },
 	{ "adaptive",		VIDIOCQCGADAPTIVE, 	VIDIOCQCSADAPTIVE,	FALSE,	flags_bool },
 	{ "equalize",		VIDIOCQCGEQUALIZE, 	VIDIOCQCSEQUALIZE,	FALSE,	flags_bool },
+	{ "userlut",		VIDIOCQCGUSERLUT, 	VIDIOCQCSUSERLUT,	FALSE,	flags_bool },
 	{ "retryerrors",	VIDIOCQCGRETRYERRORS,	VIDIOCQCSRETRYERRORS,	FALSE,	flags_bool },
 	{ "compatible", 	VIDIOCQCGCOMPATIBLE,	VIDIOCQCSCOMPATIBLE,	TRUE,	flags_compat },
 	{ "video_nr",		VIDIOCQCGVIDEONR,	VIDIOCQCSVIDEONR,	FALSE,	NULL },
@@ -222,12 +229,30 @@ void help(void) {
 "	-i	Display camera information\n"
 "	-r	Display all known registers of camera\n"
 "	-a	Display all (including empty) registers of camera\n"
-"	-b val	Set camera brightness\n"
-"	-c val	Set camera contrast\n"
-"Also module parameters are accepted, e.g. \"debug=15\"\n"
-"The module parameters can be also queried, e.g. \"debug?\"\n"
-"The parameters can be also given as a symbolic list, e.g. \"debug=logic,user\"\n"
-"The device file name (default /dev/video) can be inserted before the arguments.\n"
+"	-b val	Set brightness (camera exposure time)\n"
+"	-u val	Set hue (not supported with all sensors)\n"
+"	-o val	Set color strength (increase when setting hue)\n"
+"	-c val	Set contrast (camera gain)\n"
+"	-w val	Set whiteness (image sharpness when quality=best)\n"
+"	-g v	Enable gamma correction with the given gamma value\n"
+"	-g rg:gg:bg\n"
+"		Specify different gamma value for each color channel\n"
+"	-g rg:gg:bg:rw:gw:bw\n"
+"		Specify also white balance (0.55:0.55:0.55:1.0:1.0:1.0)\n"
+"	-g +	Enable software lookup-table\n"
+"	-e +	Same as '-g +'\n"
+"	-g -	Disable software lookup-table\n"
+"	-e -	Same as '-g -'\n"
+"	-g '?'	Display software lookup-table status and contents\n"
+"	-e '?'	Same as '-g '?''\n"
+"	-e filename.ppm\n"
+"		Generate static equalization based on image in given\n"
+"		raw PNM/PPM file.\n"
+"The \"val\" values should be between 0...65535 inclusive, 32768 is the default.\n"
+"Also module parameters are accepted, e.g. \"qcdebug=15\"\n"
+"The module parameters can be also queried, e.g. \"qcdebug?\"\n"
+"The parameters can be also given as a symbolic list, e.g. \"qcdebug=logic,user\"\n"
+"The device file name (default /dev/video0) can be inserted before the arguments.\n"
 	, progname);
 	printf("Possible module parameters and symbolic flags:\n");
 	for (i=0; i<SIZE(ioctlinfo); i++) {
@@ -295,11 +320,138 @@ void print_regs(int fd, int scanall) {
 }
 /* }}} */
 
+FILE *pnm_file = NULL;
+
+/* {{{ [fold] void pnm_nexttoken(void) */
+void pnm_nexttoken(void) {
+	int c;
+	do {
+		c = getc(pnm_file); 
+		if (c==EOF) error("unexpected end of file");
+		if (c=='#') {
+			do {
+				c = getc(pnm_file);
+				if (c==EOF) error("unexpected end of file");
+			} while (c!='\n');
+			c = ' ';
+		}
+	} while (isspace(c));
+	ungetc(c,pnm_file);
+}
+/* }}} */
+/* {{{ [fold] unsigned int pnm_readnum(void) */
+unsigned int pnm_readnum(void) {
+	unsigned int c, val = 0;
+	pnm_nexttoken();
+	do {
+		c = getc(pnm_file);
+		if (!isdigit(c)) break;
+		val = val*10 + c - '0';         /* Assume ASCII */
+	} while (1);
+	if (!isspace(c)) error("PNM file error");
+	ungetc(c,pnm_file);
+	return val;
+}
+/* }}} */
+/* {{{ [fold] void pnm_open(unsigned char *name, unsigned int *width, unsigned int *height) */
+void pnm_open(unsigned char *name, unsigned int *width, unsigned int *height) {
+	char c1,c2,c3;
+	unsigned int maxval;
+
+	pnm_file = fopen(name, "rb");
+	if (!pnm_file) error("can not open file `%s'", name);
+
+	/* Check signature */
+	c1 = getc(pnm_file);
+	c2 = getc(pnm_file);
+	c3 = getc(pnm_file);
+	if (c1!='P' || c2!='6' || !isspace(c3)) error("Not raw bits PPM file");
+	ungetc(c3,pnm_file);
+
+	/* Read dimensions */
+	*width  = pnm_readnum();
+	*height = pnm_readnum();
+	maxval  = pnm_readnum();
+	if (maxval>255) error("Does not support PNM files with more depth than 8 bits");
+	c1 = getc(pnm_file);
+	if (!isspace(c1)) error("PNM file error");
+}
+/* }}} */
+/* {{{ [fold] void pnm_close(void) */
+void pnm_close(void)
+{
+	fclose(pnm_file);
+	pnm_file = NULL;
+}
+/* }}} */
+/* {{{ [fold] void pnm_getpixel(unsigned char *r, unsigned char *g, unsigned char *b) */
+void pnm_getpixel(unsigned char *r, unsigned char *g, unsigned char *b)
+{
+	*r = getc(pnm_file);
+	*g = getc(pnm_file);
+	*b = getc(pnm_file);
+}
+/* }}} */
+
+/* {{{ [fold] void compute_lut(unsigned int (*hist)[256], unsigned char (*lut)[256]) */
+/* Compute lookup-table `lut' (256 elements) for image equalization
+ * based on given image color histogram `hist'. The LUT is also smoothed and normalized. */
+void compute_lut(unsigned int (*hist)[256], unsigned char *lut)
+{
+	static const int len = 64;			/* Smoothing kernel length, must be less than 255 */
+	static const double p = 3.0 - 1;		/* Smoothing kernel order */
+	unsigned int sum;
+	double dlut[256];
+	double min, max;
+	double norm, v, r;
+	int i,j;
+
+	/* Compute lookup-table based on the histogram */
+	sum = 0;
+	for (i=0; i<256; i++) {
+		dlut[i] = sum;
+		sum += (*hist)[i];
+	}
+
+	/* Normalize lookup-table */
+	norm = 255.0 / dlut[255];
+	for (i=0; i<256; i++) {
+		dlut[i] *= norm;
+	}
+
+	/* Compute normalization factor for smoothing kernel */
+	norm = 0.0;
+	for (j=-len; j<=len; j++) {
+		norm += pow(j+len, p) * pow(len-j, p);
+	}
+	norm = 1.0 / norm;
+
+	/* Smooth the lookup table, handle edges with point-symmetric reflections */
+	min = dlut[0];
+	max = dlut[255];
+	for (i=0; i<256; i++) {
+		r = 0.0;
+		for (j=-len; j<=len; j++) {
+			int x = i + j;
+			if (x<0) {
+				v = min - dlut[-x];
+			} else if (x>255) {
+				v = 2*max - dlut[511-x];
+			} else {
+				v = dlut[x];
+			}
+			r += v * norm * pow(j+len, p) * pow(len-j, p);
+		}
+		lut[i] = (unsigned char)CLIP(r+0.5, 0.0, 255.0);
+	}
+}
+/* }}} */
+
 /* {{{ [fold] main() */
 int main(int argc, char *argv[])
 {
 	int fd = 0,i,j,r,l,l2,c;
-	char *device = "/dev/video";
+	char *device = "/dev/video0";		/* Default should not be /dev/video which may be subdir */
 	char *argp;
 	struct video_capability vidcap;
 	struct video_window vidwin;
@@ -311,24 +463,21 @@ int main(int argc, char *argv[])
 	if (argv && argv[0] && argv[1] && argv[1][0]=='-' && argv[1][1]=='h')
 		help();
 	while (*++argv) {
-		if (argv[0][0]=='/') {
+		if (argv[0][0]=='/' || argv[0][0]=='.') {
 			/* Device filename given */
 			if (fd) close(fd);
 			device = argv[0];
 			fd = 0;
 		} else {
 			if (fd==0) {
-				/* Open the device, if it wasn't yet opened */
+				/* Open the device, if it wasn't yet opened.
+				 * Do not try to modify any settings, because if other applications have
+				 * already the camera open and are capturing, it might disturb them under 2.6.x */
 				fd = open(device,O_RDWR);	if (fd==-1) error("can not open %s",device);
 				r = ioctl(fd, VIDIOCGCAP, &vidcap);	if (r != 0) error("ioctl VIDIOCGCAP");
-				memset(&vidwin, 0, sizeof(vidwin));
-				vidwin.width = vidcap.maxwidth - 0;
-				vidwin.height = vidcap.maxheight;
-				r = ioctl(fd, VIDIOCSWIN, &vidwin);	if (r != 0) error("ioctl VIDIOCSWIN");
 				r = ioctl(fd, VIDIOCGWIN, &vidwin);	if (r != 0) error("ioctl VIDIOCGWIN");
 				vidchan.channel = 0;
 				r = ioctl(fd, VIDIOCGCHAN, &vidchan);	if (r != 0) error("ioctl VIDIOCGCHAN");
-				r = ioctl(fd, VIDIOCSCHAN, &vidchan);	if (r != 0) error("ioctl VIDIOCSCHAN");
 				r = ioctl(fd, VIDIOCGPICT, &vidpic);	if (r != 0) error("ioctl VIDIOCGPICT");
 			}
 			if (argv[0][0]=='-') {
@@ -364,6 +513,20 @@ int main(int argc, char *argv[])
 					if (r!=0) error("ioctl VIDIOCSPICT");
 					argv += c;
 					break;
+				case 'u':	/* Hue */
+					if (!argp) error("missing hue value");
+					vidpic.hue = i;
+					r = ioctl(fd, VIDIOCSPICT, &vidpic);
+					if (r!=0) error("ioctl VIDIOCSPICT");
+					argv += c;
+					break;
+				case 'o':	/* Color */
+					if (!argp) error("missing color value");
+					vidpic.colour = i;
+					r = ioctl(fd, VIDIOCSPICT, &vidpic);
+					if (r!=0) error("ioctl VIDIOCSPICT");
+					argv += c;
+					break;
 				case 'c':	/* Contrast */
 					if (!argp) error("missing contrast value");
 					vidpic.contrast = i;
@@ -371,6 +534,107 @@ int main(int argc, char *argv[])
 					if (r!=0) error("ioctl VIDIOCSPICT");
 					argv += c;
 					break;
+				case 'w':	/* Whiteness */
+					if (!argp) error("missing whiteness value");
+					vidpic.whiteness = i;
+					r = ioctl(fd, VIDIOCSPICT, &vidpic);
+					if (r!=0) error("ioctl VIDIOCSPICT");
+					argv += c;
+					break;
+				case 'g':
+				case 'e': {
+					struct qc_userlut userlut;
+					memset(&userlut, 0, sizeof(userlut));
+					if (!argp) error("missing parameter to option `-%c'", argv[0][1]);
+					switch (argp[0]) {
+					case '?':		/* Display software lut settings */
+						userlut.flags |= QC_USERLUT_VALUES;
+						r = ioctl(fd, VIDIOCQCGUSERLUT, &userlut);
+						if (r!=0) error("ioctl VIDIOCGUSERLUT");
+						printf("Software lookup-table status: %s\n", (userlut.flags & QC_USERLUT_ENABLE) ? "Enabled" : "Disabled");
+						printf("\tRed table:\n");
+						for (i=0; i<256; i++) {
+							printf("%i", userlut.lut[QC_LUT_RED + i]);
+							if (i!=255) printf(",");
+							if (((i+1) % 20)==0 || i==255) printf("\n");
+						}
+						printf("\tGreen table:\n");
+						for (i=0; i<256; i++) {
+							printf("%i", userlut.lut[QC_LUT_GREEN + i]);
+							if (i!=255) printf(",");
+							if (((i+1) % 20)==0 || i==255) printf("\n");
+						}
+						printf("\tBlue table:\n");
+						for (i=0; i<256; i++) {
+							printf("%i", userlut.lut[QC_LUT_BLUE + i]);
+							if (i!=255) printf(",");
+							if (((i+1) % 20)==0 || i==255) printf("\n");
+						}
+						break;
+					case '+':		/* Enable lookup-table */
+						userlut.flags |= QC_USERLUT_ENABLE;
+						/* Continue */
+					case '-':		/* Disable lookup-table */
+						r = ioctl(fd, VIDIOCQCSUSERLUT, &userlut);
+						if (r!=0) error("ioctl VIDIOCSUSERLUT");
+						break;
+					default:
+						switch (argv[0][1]) {
+						case 'g': {	/* Gamma */
+							double gr = 0.55;
+							double gg = 0.55;
+							double gb = 0.55;
+							double wr = 1.0;
+							double wg = 1.0;
+							double wb = 1.0;
+							r = sscanf(argp, "%lg:%lg:%lg:%lg:%lg:%lg", &gr, &gg, &gb, &wr, &wg, &wb);
+							if (r!=1 && r!=3 && r!=6) error("bad number of arguments for -g (must be 1, 3, or 6)");
+							if (r < 3) {
+								gg = gr;
+								gb = gr;
+							}
+							for (i=0; i<256; i++) {
+								userlut.lut[QC_LUT_RED   + i] = (unsigned char)CLIP(pow(i/256.0, gr)*256.0*wr+0.5, 0.0, 255.0);
+								userlut.lut[QC_LUT_GREEN + i] = (unsigned char)CLIP(pow(i/256.0, gg)*256.0*wg+0.5, 0.0, 255.0);
+								userlut.lut[QC_LUT_BLUE  + i] = (unsigned char)CLIP(pow(i/256.0, gb)*256.0*wb+0.5, 0.0, 255.0);
+							}
+							userlut.flags |= QC_USERLUT_ENABLE | QC_USERLUT_VALUES;
+							r = ioctl(fd, VIDIOCQCSUSERLUT, &userlut);
+							if (r!=0) error("ioctl VIDIOCSUSERLUT");
+						}
+						case 'e': {	/* Static equalization */
+							unsigned int width, height;
+							unsigned char r, g, b;
+							unsigned int cnt_red[256], cnt_green[256], cnt_blue[256];
+							memset(cnt_red,   0, sizeof(cnt_red));
+							memset(cnt_green, 0, sizeof(cnt_green));
+							memset(cnt_blue,  0, sizeof(cnt_blue));
+							pnm_open(argp, &width, &height);
+							/* Compute histograms for each color channel */
+							for (i=0; i<height; i++) for (j=0; j<width; j++) {
+								pnm_getpixel(&r, &g, &b);
+								cnt_red  [r]++;
+								cnt_green[g]++;
+								cnt_blue [b]++;
+							}
+							pnm_close();
+							/* Compute lookup tables based on the histograms */
+							compute_lut(&cnt_red,   &userlut.lut[QC_LUT_RED]);
+							compute_lut(&cnt_green, &userlut.lut[QC_LUT_GREEN]);
+							compute_lut(&cnt_blue,  &userlut.lut[QC_LUT_BLUE]);
+							/* Send the new lookup table to the driver */
+							userlut.flags |= QC_USERLUT_ENABLE | QC_USERLUT_VALUES;
+							r = ioctl(fd, VIDIOCQCSUSERLUT, &userlut);
+							if (r!=0) error("ioctl VIDIOCSUSERLUT");
+							break;
+						}
+						default:
+							error("this can't happen");
+						}
+					}
+					argv += c;
+					break;
+				}
 				default:
 					error("invalid option");
 				}
@@ -424,9 +688,9 @@ int main(int argc, char *argv[])
 					}
 					r = ioctl(fd, ioctlinfo[i].set, &c);	if (r != 0) error("ioctl set %s", ioctlinfo[i].name);
 					printf("%i\n", c);
-				} else
+				} else {
 					error("invalid configuration request type");
-				
+				}
 			}
 		}
 	}
