@@ -1,5 +1,79 @@
 #ifndef _LINUX_QUICKCAM_H
 #define _LINUX_QUICKCAM_H
+
+/* {{{ [fold] SECTION: common definitions with userspace applications */
+#define QC_LUT_SIZE	(3*256)		/* Lookup table definition for equalization */
+#define QC_LUT_RED	0
+#define QC_LUT_GREEN	256
+#define QC_LUT_BLUE	512
+
+struct qc_userlut {
+	unsigned int flags;
+#define QC_USERLUT_DEFAULT 1		/* If set, change default settings or the current camera otherwise */
+#define QC_USERLUT_ENABLE  2		/* If set, enable user-specified LUT, or otherwise disable it */
+#define QC_USERLUT_VALUES  4		/* Load new (or store old) values into the lookup-table */
+	unsigned char lut[QC_LUT_SIZE];	/* Lookup table to set or read */
+};
+
+#define VIDEO_PALETTE_BAYER	(('q'<<8) | 1)	/* Grab video in raw Bayer format */
+#define VIDEO_PALETTE_MJPEG	(('q'<<8) | 2)	/* Grab video in compressed MJPEG format */
+#define VID_HARDWARE_QCAM_USB	(('q'<<8) | 50)	/* Device type */
+
+/* Private IOCTL calls */
+#define QC_IOCTLBASE		220				/* Don't use same numbers as Philips driver */
+#define VIDIOCQCGDEBUG		_IOR ('v',QC_IOCTLBASE+0, int)	/* Gets the debug output, bitfield */
+#define VIDIOCQCSDEBUG		_IOWR('v',QC_IOCTLBASE+0, int)	/* Sets the debug output, bitfield */
+#define VIDIOCQCGKEEPSETTINGS	_IOR ('v',QC_IOCTLBASE+1, int)	/* Get keep picture settings across one open to another (0-1) */
+#define VIDIOCQCSKEEPSETTINGS	_IOWR('v',QC_IOCTLBASE+1, int)	/* Set keep picture settings across one open to another (0-1) */
+#define VIDIOCQCGSETTLE		_IOR ('v',QC_IOCTLBASE+2, int)	/* Get if we let image brightness settle (0-1) */
+#define VIDIOCQCSSETTLE		_IOWR('v',QC_IOCTLBASE+2, int)	/* Set if we let image brightness settle (0-1) */
+#define VIDIOCQCGSUBSAMPLE	_IOR ('v',QC_IOCTLBASE+3, int)	/* Gets the speed (0-1) */
+#define VIDIOCQCSSUBSAMPLE	_IOWR('v',QC_IOCTLBASE+3, int)	/* Sets the speed (0-1) */
+#define VIDIOCQCGCOMPRESS	_IOR ('v',QC_IOCTLBASE+4, int)	/* Gets the compression mode (0-1) */
+#define VIDIOCQCSCOMPRESS	_IOWR('v',QC_IOCTLBASE+4, int)	/* Sets the compression mode (0-1) */
+#define VIDIOCQCGFRAMESKIP	_IOR ('v',QC_IOCTLBASE+5, int)	/* Get frame capture frequency (0-10) */
+#define VIDIOCQCSFRAMESKIP	_IOWR('v',QC_IOCTLBASE+5, int)	/* Set frame capture frequency (0-10) */
+#define VIDIOCQCGQUALITY	_IOR ('v',QC_IOCTLBASE+6, int)	/* Gets the interpolation mode (0-2) */
+#define VIDIOCQCSQUALITY	_IOWR('v',QC_IOCTLBASE+6, int)	/* Sets the interpolation mode (0-2) */
+#define VIDIOCQCGADAPTIVE	_IOR ('v',QC_IOCTLBASE+7, int)	/* Get automatic adaptive brightness control (0-1) */
+#define VIDIOCQCSADAPTIVE	_IOWR('v',QC_IOCTLBASE+7, int)	/* Set automatic adaptive brightness control (0-1) */
+#define VIDIOCQCGEQUALIZE	_IOR ('v',QC_IOCTLBASE+8, int)	/* Get equalize image (0-1) */
+#define VIDIOCQCSEQUALIZE	_IOWR('v',QC_IOCTLBASE+8, int)	/* Set equalize image (0-1) */
+#define VIDIOCQCGRETRYERRORS	_IOR ('v',QC_IOCTLBASE+9, int)	/* Get if we retry when capture fails (0-1) */
+#define VIDIOCQCSRETRYERRORS	_IOWR('v',QC_IOCTLBASE+9, int)	/* Set if we retry when capture fails (0-1) */
+#define VIDIOCQCGCOMPATIBLE	_IOR ('v',QC_IOCTLBASE+10,int)	/* Get enable workaround for bugs, bitfield */
+#define VIDIOCQCSCOMPATIBLE	_IOWR('v',QC_IOCTLBASE+10,int)	/* Set enable workaround for bugs, bitfield */
+#define VIDIOCQCGVIDEONR	_IOR ('v',QC_IOCTLBASE+11,int)	/* Get videodevice number (/dev/videoX) */
+#define VIDIOCQCSVIDEONR	_IOWR('v',QC_IOCTLBASE+11,int)	/* Set videodevice number (/dev/videoX) */
+#define VIDIOCQCGUSERLUT	_IOR ('v',QC_IOCTLBASE+12,struct qc_userlut)	/* Get user-specified lookup-table correction */
+#define VIDIOCQCSUSERLUT	_IOWR('v',QC_IOCTLBASE+12,struct qc_userlut)	/* Set user-specified lookup-table correction */
+
+#define VIDIOCQCGSTV		_IOWR('v',QC_IOCTLBASE+20,int)	/* Read STV chip register */
+#define VIDIOCQCSSTV		_IOW ('v',QC_IOCTLBASE+20,int)	/* Write STV chip register */
+#define VIDIOCQCGI2C		_IOWR('v',QC_IOCTLBASE+21,int)	/* Read I2C chip register */
+#define VIDIOCQCSI2C		_IOW ('v',QC_IOCTLBASE+21,int)	/* Write I2C chip register */
+
+/* Debugging message choices */
+#define QC_DEBUGUSER		(1<<0)	/* Messages for interaction with user space (system calls) */
+#define QC_DEBUGCAMERA		(1<<1)	/* Messages for interaction with the camera */
+#define QC_DEBUGINIT		(1<<2)	/* Messages for each submodule initialization/deinit */
+#define QC_DEBUGLOGIC		(1<<3)	/* Messages for entering and failing important functions */
+#define QC_DEBUGERRORS		(1<<4)	/* Messages for all error conditions */
+#define QC_DEBUGADAPTATION	(1<<5)	/* Messages for automatic exposure control workings */
+#define QC_DEBUGCONTROLURBS	(1<<6)	/* Messages for sending I2C control messages via USB */
+#define QC_DEBUGBITSTREAM	(1<<7)	/* Messages for finding chunk codes from camera bitstream */
+#define QC_DEBUGINTERRUPTS	(1<<8)	/* Messages for each interrupt */
+#define QC_DEBUGMUTEX		(1<<9)	/* Messages for acquiring/releasing the mutex */
+#define QC_DEBUGCOMMON		(1<<10)	/* Messages for some common warnings */
+#define QC_DEBUGFRAME		(1<<11)	/* Messages related to producer-consumer in qc_frame_* functions */
+#define QC_DEBUGALL		(~0)	/* Messages for everything */
+
+/* Compatibility choices */
+#define QC_COMPAT_16X		(1<<0)
+#define QC_COMPAT_DBLBUF	(1<<1)
+#define QC_COMPAT_TORGB		(1<<2)	/* Video4Linux API is buggy and doesn't specify byte order for RGB images */
+/* }}} */
+
 #ifdef __KERNEL__
 
 #include <linux/config.h>
@@ -21,24 +95,29 @@
 #include <linux/videodev.h>
 #include <linux/usb.h>
 #include <asm/pgtable.h>		/* This is required for testing pte_offset_map */
+#include <linux/spinlock.h>
 
 /* {{{ [fold] SECTION: user configuration */
-#define VERSION		"QuickCam USB $Date: 2003/12/31 16:15:06 $"
+#define VERSION		"QuickCam USB $Date: 2004/07/29 18:12:39 $"
 #ifndef COMPRESS
 #define COMPRESS	1		/* 1=include compression support, 0=otherwise */
 #endif
 #ifndef DEBUGLEVEL
 #define DEBUGLEVEL	QC_DEBUGCOMMON
 #endif
-//#define NDEBUG			/* Disable all debug code */
-#ifndef NDEBUG
+#ifdef NDEBUG				/* Enable debugging if DEBUG is defined; if (also) NDEBUG is defined, disable debugging */
+#undef DEBUG
+#endif
+//#define DEBUG				/* Enable debug code */
+#ifdef DEBUG
 #define PARANOID	1		/* Check consistency of driver state */
 #else
 #define PARANOID	0
 #endif
-#define MEASURE_ADAPT_DELAY 0		/* Measure adaptation delay, only for test purposes */
 /* Default (initial) values */
 #define DEFAULT_BGR	TRUE		/* Use BGR byte order by default (and torgb converts to RGB)? */
+
+#define DUMPDATA	0		/* Dump data from camera to user, no conversion nor length checks (see show.c) */
 /* }}} */
 /* {{{ [fold] SECTION: general utility definitions and macros */
 #define FALSE			0
@@ -72,8 +151,8 @@ typedef unsigned char Bool;
 #warning "Unsupported kernel, may or may not work. Good luck!"
 #endif
 
-#if LINUX_VERSION_CODE>=KERNEL_VERSION(2,4,0) && defined(CONFIG_PROC_FS) && defined(CONFIG_VIDEO_PROC_FS)
-#define HAVE_PROCFS 1
+#if LINUX_VERSION_CODE>=KERNEL_VERSION(2,4,0) && defined(CONFIG_PROC_FS)
+#define HAVE_PROCFS 1		/* FIXME: I don't think there's any reason to disable procfs with 2.2.x */
 #else
 #define HAVE_PROCFS 0
 #warning "procfs support disabled"
@@ -158,7 +237,8 @@ static inline struct urb *qc_usb_alloc_urb(int packets) { return usb_alloc_urb(p
 /* {{{ [fold] SECTION: debugging */
 #undef PDEBUG			/* undef it, just in case */
 #define POISON_VAL	0x5B
-#ifndef NDEBUG
+
+#ifdef DEBUG
 
 #include <linux/time.h>
 /* PDEBUG is rather heavyweight macro and should be used only for debugging--not for general messages */
@@ -184,6 +264,10 @@ static inline struct urb *qc_usb_alloc_urb(int packets) { return usb_alloc_urb(p
 	(x).idebug_var = NULL; \
 	POISON(x); \
 } while(0)
+#define IDEBUG_EXIT_NOPOISON(x)	do { \
+	IDEBUG_TEST(x); \
+	(x).idebug_var = NULL; \
+} while(0)
 #define TEST_BUG(condition) \
 	do { \
 		if ((condition)!=0) { \
@@ -223,13 +307,14 @@ static inline struct urb *qc_usb_alloc_urb(int packets) { return usb_alloc_urb(p
 #define IDEBUG_INIT(x)
 #define IDEBUG_TEST(x)
 #define IDEBUG_EXIT(x)
+#define IDEBUG_EXIT_NOPOISON(x)
 #define TEST_BUG(x)
 #define TEST_BUGR(x)
 #define TEST_BUG_MSG(cond, fmt, args...)
 #define TEST_BUGR_MSG(cond, fmt, args...)
 #define POISON(obj)
 
-#endif /* !NDEBUG */
+#endif /* DEBUG */
 
 //gcc is buggy? This doesn't work
 //#define PRINTK(lvl, fmt, args...)	printk(lvl "quickcam: " fmt "\n", ## args)
@@ -321,11 +406,14 @@ struct qc_frame_data {
 	int maxrawdatalen;		/* Maximum amount of data we are willing to accept in bytes, */
 					/* zero indicates that we are not grabbing current frame (but just throwing data away) */
 	volatile unsigned int head;	/* The buffer to be captured next (empty or grabbing, if full, then whole buffer is full) */
-	unsigned int tail;		/* The buffer to be consumed next (full, unless equals head, then it is empty/grabbing) */
+	volatile unsigned int tail;	/* The buffer to be consumed next (full, unless equals head, then it is empty/grabbing) */
+	spinlock_t tail_lock;		/* Protects tail changes */
+	volatile Bool tail_in_use;	/* TRUE, when consumer is processing the frame pointed to by tail */
+
 	wait_queue_head_t wq;		/* Processes waiting for more data in the buffer */
 	volatile int waiting;		/* Number of processes waiting in the wait queues */
 	volatile Bool exiting;		/* Set to TRUE when we want to quit */
-	int lost_frames;		/* Increased by one for every lost (non-captured by applications) frame, reset when a frame is captured */
+	volatile int lost_frames;	/* Increased by one for every lost (non-captured by applications) frame, reset when a frame is captured */
 	IDEBUG_VAR
 };
 /* }}} */
@@ -346,12 +434,9 @@ struct qc_mjpeg_data {
 };
 /* }}} */
 /* {{{ [fold] qc_fmt_data: Format conversion routines private data */
-#define LUT_SIZE	(3*256)		/* Lookup table definition for equalization */
-#define LUT_RED		0
-#define LUT_GREEN	256
-#define LUT_BLUE	512
 struct qc_fmt_data {
-	unsigned char lut[LUT_SIZE];
+	unsigned char userlut[QC_LUT_SIZE];	/* User specified fixed look-up table, initialized when camera is plugged in */
+	unsigned char lut[QC_LUT_SIZE];		/* Dynamically calculated LUT, for which userlut is applied to */
 #if COMPRESS
 	struct qc_mjpeg_data mjpeg_data;
 	Bool compress;			/* Was compression subsystem initialized? */
@@ -387,17 +472,19 @@ struct qc_adapt_data {
 /* {{{ [fold] qc_settings_data: User settings given by qcset or module parameters, initialized when camera is plugged in */
 struct qc_settings_data {
 	unsigned int keepsettings  : 1;	/* Keep all settings at each camera open (or reset most of them) */
-	unsigned int settle        : 1;	/* Before giving first picture after open, let brightness settle */
 	unsigned int subsample     : 1;	/* Normal or sub-sample (sub-sample to increase the speed) */
 	unsigned int compress      : 1;	/* Enable compressed mode if available (higher framerate) */
 	unsigned int frameskip     : 4;	/* How many frames to skip (higher=lower framerate) */
-	unsigned int quality       : 2;	/* Quality of format conversion (higher=better but slower) */
+	unsigned int quality       : 3;	/* Quality of format conversion (higher=better but slower) */
 	unsigned int adaptive      : 1;	/* Use automatic exposure control */
 	unsigned int equalize      : 1;	/* Equalize images */
+	unsigned int userlut       : 1;	/* Apply user-specified lookup-table */
 	unsigned int retryerrors   : 1;	/* If errors happen when capturing an image, retry a few times? */
 	unsigned int compat_16x    : 1;	/* Compatibility: force image size to multiple of 16 */
 	unsigned int compat_dblbuf : 1;	/* Compatibility: fake doublebuffering for applications */
 	unsigned int compat_torgb  : 1;	/* Compatibility: use RGB data order, not BGR */
+	unsigned int settle        : 8;	/* Maximum number of frames to wait image brightness to settle */
+	/* Total: 25 bits */
 };
 /* }}} */
 
@@ -407,8 +494,9 @@ struct quickcam {
 	struct semaphore lock;			/* Allow only one process to access quickcam at a time */
 	struct list_head list;			/* All cameras are in a doubly linked list */
 	int users;				/* User count (simultaneous open count) */
-	struct usb_device *dev;			/* USB device, set to NULL when camera disconnected while device open */
+	struct usb_device *dev;			/* USB device, set to NULL when camera disconnected and interrupts disabled */
 	unsigned char iface;			/* The interface number in the camera device we are bound to */
+	Bool connected;				/* Set to FALSE immediately when the camera is disconnected (even before interrupts are disabled) */
 	struct video_device vdev;		/* Used for registering the camera driver for Video4Linux */
 	struct qc_settings_data settings;	/* General user settings set with e.g. qcset */
 #if HAVE_PROCFS
@@ -416,6 +504,7 @@ struct quickcam {
 #endif
 	/* The following entries are initialized in qc_v4l_init() when the camera device is opened */
 	struct video_picture vpic;		/* Contains the last values set by user (which is reported to user) */
+	Bool vpic_pending;			/* Settings in vpic were changed but are not yet in effect */
 	struct video_window vwin;		/* Contains the image size and position the user is expecting */
 
 	/* Private structures for each module, initialized in corresponding init-function */
@@ -426,7 +515,7 @@ struct quickcam {
 	struct qc_frame_data frame_data;	/* Filled when the device is opened */
 	struct qc_capt_data capt_data;		/* Filled when the device is opened */
 	struct qc_isoc_data isoc_data;		/* Filled when the device is opened */
-	struct qc_fmt_data fmt_data;		/* Filled when the device is opened */
+	struct qc_fmt_data fmt_data;		/* Mostly filled when the device is opened, except for userlut */
 
 	u8 dmabuf[35];				/* Temporary buffer which may be used for DMA transfer */
 };
@@ -466,6 +555,7 @@ int qc_i2c_setw(struct quickcam *qc, unsigned char reg, unsigned short val);
 
 int qc_stv_set(struct quickcam *qc, unsigned short reg, unsigned char val);
 int qc_stv_setw(struct quickcam *qc, unsigned short reg, unsigned short val);
+int qc_stv_get(struct quickcam *qc, unsigned short reg);
 
 /* Image format conversion routines */
 int qc_fmt_convert(struct quickcam *qc, unsigned char *src, unsigned int src_len, unsigned char *dst, unsigned int dst_len, int *midvalue);
@@ -484,64 +574,9 @@ int qc_get_i2c(struct quickcam *qc, const struct qc_sensor *sensor, int reg);
 void qc_frame_flush(struct quickcam *qc);
 
 void qc_usleep(unsigned long usec);
-extern int debug;				/* Driver debuglevel */
+extern int qcdebug;				/* Driver debuglevel */
 /* }}} */
+
 #endif /* __KERNEL__ */
-/* {{{ [fold] SECTION: common definitions with userspace applications */
-#define VIDEO_PALETTE_BAYER	(('q'<<8) | 1)	/* Grab video in raw Bayer format */
-#define VIDEO_PALETTE_MJPEG	(('q'<<8) | 2)	/* Grab video in compressed MJPEG format */
-#define VID_HARDWARE_QCAM_USB	(('q'<<8) | 50)	/* Device type */
-
-/* Private IOCTL calls */
-#define QC_IOCTLBASE		220				/* Don't use same numbers as Philips driver */
-#define VIDIOCQCGDEBUG		_IOR ('v',QC_IOCTLBASE+0, int)	/* Gets the debug output, bitfield */
-#define VIDIOCQCSDEBUG		_IOWR('v',QC_IOCTLBASE+0, int)	/* Sets the debug output, bitfield */
-#define VIDIOCQCGKEEPSETTINGS	_IOR ('v',QC_IOCTLBASE+1, int)	/* Get keep picture settings across one open to another (0-1) */
-#define VIDIOCQCSKEEPSETTINGS	_IOWR('v',QC_IOCTLBASE+1, int)	/* Set keep picture settings across one open to another (0-1) */
-#define VIDIOCQCGSETTLE		_IOR ('v',QC_IOCTLBASE+2, int)	/* Get if we let image brightness settle (0-1) */
-#define VIDIOCQCSSETTLE		_IOWR('v',QC_IOCTLBASE+2, int)	/* Set if we let image brightness settle (0-1) */
-#define VIDIOCQCGSUBSAMPLE	_IOR ('v',QC_IOCTLBASE+3, int)	/* Gets the speed (0-1) */
-#define VIDIOCQCSSUBSAMPLE	_IOWR('v',QC_IOCTLBASE+3, int)	/* Sets the speed (0-1) */
-#define VIDIOCQCGCOMPRESS	_IOR ('v',QC_IOCTLBASE+4, int)	/* Gets the compression mode (0-1) */
-#define VIDIOCQCSCOMPRESS	_IOWR('v',QC_IOCTLBASE+4, int)	/* Sets the compression mode (0-1) */
-#define VIDIOCQCGFRAMESKIP	_IOR ('v',QC_IOCTLBASE+5, int)	/* Get frame capture frequency (0-10) */
-#define VIDIOCQCSFRAMESKIP	_IOWR('v',QC_IOCTLBASE+5, int)	/* Set frame capture frequency (0-10) */
-#define VIDIOCQCGQUALITY	_IOR ('v',QC_IOCTLBASE+6, int)	/* Gets the interpolation mode (0-2) */
-#define VIDIOCQCSQUALITY	_IOWR('v',QC_IOCTLBASE+6, int)	/* Sets the interpolation mode (0-2) */
-#define VIDIOCQCGADAPTIVE	_IOR ('v',QC_IOCTLBASE+7, int)	/* Get automatic adaptive brightness control (0-1) */
-#define VIDIOCQCSADAPTIVE	_IOWR('v',QC_IOCTLBASE+7, int)	/* Set automatic adaptive brightness control (0-1) */
-#define VIDIOCQCGEQUALIZE	_IOR ('v',QC_IOCTLBASE+8, int)	/* Get equalize image (0-1) */
-#define VIDIOCQCSEQUALIZE	_IOWR('v',QC_IOCTLBASE+8, int)	/* Set equalize image (0-1) */
-#define VIDIOCQCGRETRYERRORS	_IOR ('v',QC_IOCTLBASE+9, int)	/* Get if we retry when capture fails (0-1) */
-#define VIDIOCQCSRETRYERRORS	_IOWR('v',QC_IOCTLBASE+9, int)	/* Set if we retry when capture fails (0-1) */
-#define VIDIOCQCGCOMPATIBLE	_IOR ('v',QC_IOCTLBASE+10,int)	/* Get enable workaround for bugs, bitfield */
-#define VIDIOCQCSCOMPATIBLE	_IOWR('v',QC_IOCTLBASE+10,int)	/* Set enable workaround for bugs, bitfield */
-#define VIDIOCQCGVIDEONR	_IOR ('v',QC_IOCTLBASE+11,int)	/* Get videodevice number (/dev/videoX) */
-#define VIDIOCQCSVIDEONR	_IOWR('v',QC_IOCTLBASE+11,int)	/* Set videodevice number (/dev/videoX) */
-
-#define VIDIOCQCGSTV		_IOWR('v',QC_IOCTLBASE+20,int)	/* Read STV chip register */
-#define VIDIOCQCSSTV		_IOW ('v',QC_IOCTLBASE+20,int)	/* Write STV chip register */
-#define VIDIOCQCGI2C		_IOWR('v',QC_IOCTLBASE+21,int)	/* Read I2C chip register */
-#define VIDIOCQCSI2C		_IOW ('v',QC_IOCTLBASE+21,int)	/* Write I2C chip register */
-
-/* Debugging message choices */
-#define QC_DEBUGUSER		(1<<0)	/* Messages for interaction with user space (system calls) */
-#define QC_DEBUGCAMERA		(1<<1)	/* Messages for interaction with the camera */
-#define QC_DEBUGINIT		(1<<2)	/* Messages for each submodule initialization/deinit */
-#define QC_DEBUGLOGIC		(1<<3)	/* Messages for entering and failing important functions */
-#define QC_DEBUGERRORS		(1<<4)	/* Messages for all error conditions */
-#define QC_DEBUGADAPTATION	(1<<5)	/* Messages for automatic exposure control workings */
-#define QC_DEBUGCONTROLURBS	(1<<6)	/* Messages for sending I2C control messages via USB */
-#define QC_DEBUGBITSTREAM	(1<<7)	/* Messages for finding chunk codes from camera bitstream */
-#define QC_DEBUGINTERRUPTS	(1<<8)	/* Messages for each interrupt */
-#define QC_DEBUGMUTEX		(1<<9)	/* Messages for acquiring/releasing the mutex */
-#define QC_DEBUGCOMMON		(1<<10)	/* Messages for some common warnings */
-#define QC_DEBUGALL		(~0)	/* Messages for everything */
-
-/* Compatibility choices */
-#define QC_COMPAT_16X		(1<<0)
-#define QC_COMPAT_DBLBUF	(1<<1)
-#define QC_COMPAT_TORGB		(1<<2)	/* Video4Linux API is buggy and doesn't specify byte order for RGB images */
-/* }}} */
 
 #endif /* __LINUX_QUICKCAM_H */
